Coarray Fortran Tutorial 
========================

## Overview

This directory contains the following standalone parallel Fortran 2018 programs:

* [hi.f90] : A minimal "Hello, world!" program that identifies parallel images
* [hello.f90] : A "Hello, world!" program with communication and synchronization
* [async-hello.f90] : An asynchronous "Hello, world!" program that demonstrates events
* [heat-equation.f90] : An unsteady 2D heat equation solver

These programs demonstrate Fortran's parallel features commonly referred to as
"Coarray Fortran."

### Prerequisites
* The [GCC], NAG, HPE Cray, or Intel Fortran compilers
* _Only if using [GCC]_: The [OpenCoarrays] compiler wrapper (`caf`) and program
  launcher (`cafrun`)

### Basic usage with Make

Given a working parallel Fortran compiler, you may be able to build and run
all the examples with the following command:

```
make all run
```

You can build and run a particular example with the make target `run-<example>`, eg:
```
make run-hi
```

The settings at the top of the Makefile can be used to customize for your particular
system, for example to request parallel execution with 16 processes:

```
make all run PROCS=16
```

Or on a Cray system selecting the `ftn` compiler:

```
make all run CAF=ftn
```

Alternatively you can invoke compile and run commands directly as appropriate for
your system, as explained in the following section.


## Heat Equation Solver

The numerical algorithm uses 2nd-order-accurate central finite differencing
in space and 1st-order-accurate explicit Euler advancement in time.

### Parallel execution with GCC and OpenCoarrays

With the [GCC] Fortran compiler (`gfortran`) and the
[OpenCoarrays] parallel runtime library installed, compile this program
as a standalone file and run it as follows:

```
caf -o heat-equation heat-equation.f90
cafrun -n 2 ./heat-equation
```

where you may replace `2` in the above line with the desired number of images.

### Parallel execution with the Intel `ifx` compiler

With the Intel `ifx` Fortran compiler installed, 

```
ifx -o heat-equation -coarray heat-equation.f90 
export FOR_COARRAY_NUM_IMAGES=2
./heat-equation
```

### Parallel execution with the Cray Fortran compiler (`ftn`) on Perlmutter

This Cray compiler can compile the two `*hello.f90` programs.
A compiler bug prevents the compilation of the heat equation 
solver.  Execute the following commands:

```
module load PrgEnv-cray
ftn -o async-hello async-hello.f90 
salloc -N1 -t60 -Am2878 -C cpu -q interactive
srun -n32 async-hello
```

where the `salloc` command requests one node for interactive use and the
`srun` command launches the compiled `async-hello` program in 32 images.

### Serial execution with `gfortran` *without* requiring OpenCoarrays

Execute

```
gfortran -o heat-equation -fcoarray=single heat-equation.f90
./heat-equation
```

## Heat Equation Exercise

In addition to demonstrating parallel features of Fortran 2018, this example
shows an object-oriented, functional programming style based on Fortran's
user-defined operators such as the `.laplacian.` operator defined in this
example.  To demonstrate the expressive power and flexibility of this
approach, try modifying the main program to use 2nd-order
Runge-Kutta time advancement:
```
T_half = T + 0.5*dt*alpha* .laplacian. T
call T%exchange_halo
sync all
T = T + dt*alpha* .laplacian. T_half
call T%exchange_halo
sync all
```
You'll need to append `, T_half` to the declaration `type(subdomain_2D_t) T`.
With some care, you could modify the main program to use any desired order of
Runge-Kutta algorithm without changing any of the supporting code.

This example also demonstrates a benefit of Fortran's facility for declaring a
procedure to be `pure`: the semantics of `pure` procedures essentially
guarantees that the above right-hand-side expressions can be evaluated fully
asynchronously across all images.  No operator can modify state that would be
observable by another operator other than via the first operator's result. This
would be true even if an operator executing on one image performs communication
to _get_ data from another image via a coarray.  To reduce communication waiting
times, however, each image in our example proactively _puts_ data onto
neighboring images.  Puts generally outperform gets because the data can be
shipped off as soon the data are ready.  With the exception of one coarray
allocation in the `define` procedure, all procedures are asynchronous and all
image control is exposed in the main program.

## Asynchronous Hello World Exercise

Try adjusting the `delay_magnitude` constant to larger or smaller non-negative 
values.  For each new value, recompile once and rerun the program multiple times.
Explain the resulting program output.

---
The files [hi.f90], [hello.f90] and [async-hello.f90] are subject to 
[the licensing terms in LICENSE-hello.txt](LICENSE-hello.txt).    
All other files in this directory are subject to 
[the licensing terms in LICENSE.txt](LICENSE.txt)

[heat-equation.f90]: ./heat-equation.f90
[hello.f90]: ./hello.f90
[hi.f90]: ./hi.f90
[async-hello.f90]: ./async-hello.f90
[GCC]: https://gcc.gnu.org
[OpenCoarrays]: https://github.com/sourceryinstitute/opencoarrays
